// Copyleft (C) 2003 pawfal entertainment PLC
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

#include "dada.h"
#include "deque"
#include "map"
#include "vector"
#include "string"
#include "SceneGraph.h"

#ifndef N_RENDERER
#define N_RENDERER

using namespace std;

namespace fluxus
{

class State;
class Primitive;

class Renderer
{
public:
	Renderer();
	~Renderer();
	
	// retained mode
	void BeginScene(bool PickMode=false);
	void Render();
	void EndScene();	
	int  AddPrimitive(Primitive *Prim);
	Primitive *GetPrimitive(int ID);
	// make sure it's removed from the physics engine too
	void RemovePrimitive(int ID);
	// moves the primitive (and it's subtree) from it's position in the
	// graph and adds it to the root.
	void DetachPrimitive(int ID);
	dMatrix GetGlobalTransform(int ID);
	dBoundingBox GetBoundingBox(int ID);
	void Clear();
	void DrawAxes();
	void DrawText(const string &Text);
	
	// immediate mode (don't delete prim before frame end - it's immediate-ish :))
	void RenderPrimitive(Primitive *Prim);
	
	// state accessors
	State *GetState();
	void ApplyState();
	void PushState();
	void PopState();
	
	// grabs the object state and makes it current
	void Grab(int ID);
	void UnGrab();
	Primitive *Grabbed() { return m_Grabbed; }
	
	int Select(int x, int y, int size);
	
	// applies the transform of a primitive to it's geometry
	void Apply(int ID);
	
	void SetMotionBlur(bool s, float a=0.02) { m_MotionBlur=s; m_Fade=a; }
	void SetResolution(int x, int y) { m_Width=x; m_Height=y; m_Initialised=false; }
	dMatrix *GetCamera() { return &m_Camera; }
	void LockCamera(int Prim);
	void SetOrtho(bool s) { m_Ortho=s; m_Initialised=false; }
	void SetFrustum(float u, float d, float l, float r, float f, float b) 
		{ m_Up=u; m_Down=d; m_Left=l; m_Right=r; m_Front=f; m_Back=b; m_Initialised=false; }
	unsigned int LoadTexture(const string &Filename) { return m_World.LoadTexture(Filename); }
	void ShowAxis(bool s) { m_ShowAxis=s; }
	void PrintSceneGraph() { m_World.Dump(); }
	void SetBGColour(const dColour &s) { m_BGColour=s; }
	void SetClearFrame(bool s) { m_ClearFrame=s; }
	void SetBackFaceCull(bool s) { m_BackFaceCull=s; m_Initialised=false; }
	
	typedef void (cb)();
	void SetEngineCallback(cb *s) { EngineCallback=s; }
	
private:
	bool  m_Initialised;
	int   m_Width,m_Height;
	bool  m_MotionBlur;
	float m_Fade;
	bool  m_InScene;
	bool  m_Ortho;
	int   m_CameraAttached;
	bool  m_LockedCamera;
	bool  m_ShowAxis;
	Primitive *m_Grabbed;
	dColour m_BGColour;
	bool m_ClearFrame;
	float m_Up,m_Down,m_Left,m_Right,m_Front,m_Back;
	bool m_BackFaceCull;
	
	dMatrix m_Camera;
    deque<State> m_StateStack;
    SceneGraph m_World;

    /////////////////////////////////////////////////////
    // Immediate-ish mode
    // we need to keep a record of immediate mode requests to
    // potentially save in the flx file. as primitive pointers
    // are stored, prims drawn in immediate mode have to stick
    // around till the end of frame at least.
    struct IMItem
    {
    	State m_State;
    	Primitive *m_Primitive;
   	};
	
   	vector<IMItem*> m_IMRecord;
    // also renders the IM stuff in one batch, as they could have been
    // loaded from an flx file
    void RenderIMPrimitives();
    void ClearIMPrimitives();

    // need to know, so we can delete the IM primitives newed by the loader.
    bool m_LoadedFromFlx;

	// info for picking mode
	struct SelectInfo
	{
		int x,y;
		int size;
	};
	
	SelectInfo m_SelectInfo;
	
    cb *EngineCallback;

    friend istream &operator>>(istream &s, Renderer &o);
	friend ostream &operator<<(ostream &s, Renderer &o);
};

istream &operator>>(istream &s, Renderer &o);
ostream &operator<<(ostream &s, Renderer &o);
	
};

#endif
